IMPLEMENTATION MODULE Liste7;
(* Bibliotheksmodul zur Operation auf Listen mit beliebigen
   Datenstrukturen im Informationsteil.

   Autor:      Michael Broser
   Version:    1.2        vom  23.4.89
   System:     LOGITECH Modula-2 Version 3.0         *)

FROM SYSTEM IMPORT ADDRESS;

FROM Storage IMPORT ALLOCATE, DEALLOCATE;


(************************** TYPENDEKLARATION  ***********************)

TYPE tZeiger = POINTER TO eintrag;

     eintrag = RECORD
                  info: ADDRESS;
                  vorgaenger, nachfolger: tZeiger;
               END;

     tList   = POINTER TO RECORD
                            anker, aktuell: tZeiger
                          END;


(************************  INTERNE HILFSPROZEDUREN  ****************)

PROCEDURE ErzeugeNeuknoten(VAR neu: tZeiger; Adr: ADDRESS);
    BEGIN
      NEW(neu);
      neu^.info := Adr;
    END ErzeugeNeuknoten;


PROCEDURE nacheinfuegen(VAR neu,alt :tZeiger);
   VAR hilf : tZeiger;
   BEGIN
      hilf := alt^.nachfolger;
      neu^.nachfolger:= hilf;
      alt^.nachfolger:= neu;
      neu^.vorgaenger:= alt;
      hilf^.vorgaenger := neu;
   END nacheinfuegen;

PROCEDURE voreinfuegen(VAR neu,alt : tZeiger);
   VAR hilf : tZeiger;
   BEGIN
      hilf := alt^.vorgaenger;
      neu^.vorgaenger := alt^.vorgaenger;
      alt^.vorgaenger := neu;
      neu^.nachfolger := alt;
      hilf^.nachfolger := neu;
   END voreinfuegen;

(*******************************************************************)

PROCEDURE ErzeugeListe(VAR Liste:tList);
    (* Initialisiert die Liste. Diese Prozedur mu als Erstes aufge-
       rufen werden. *)
    BEGIN
      NEW(Liste);
      WITH Liste^ DO
            NEW(anker);
            anker^.nachfolger := anker;
            anker^.vorgaenger := anker;
            anker^.info       := NIL;
            aktuell := anker;
         END; (* of WITH *)
   END ErzeugeListe;


(*******************  Einfuegeoperationen in die Liste  ************)


PROCEDURE EinfuegenKopf(VAR Liste:tList; Adr:ADDRESS);
    (* Fgt einen neuen Datensatz am Kopf der Liste ein. *)
  VAR
      neu : tZeiger;
  BEGIN
         ErzeugeNeuknoten(neu,Adr);
         nacheinfuegen(neu,Liste^.anker);
  END EinfuegenKopf;


PROCEDURE EinfuegenSchwanz(VAR Liste:tList; Adr:ADDRESS);
    (* Fgt einen neuen Datensatz am Schwanz der Liste ein. *)
  VAR
      neu: tZeiger;
  BEGIN
         ErzeugeNeuknoten(neu,Adr);
         voreinfuegen(neu,Liste^.anker);
  END EinfuegenSchwanz;


PROCEDURE SortiertEinfuegen(VAR Liste:tList;Adr:ADDRESS;
                            Vergleich: VergleicheProc);
    (* Fgt einen neuen Datensatz sortiert in die Liste ein. Im Parameter
       Vergleich wird die Prozedur fr den Vergleich zweier Listenelemente
       bergeben. Das einzufgende Element ist bei der Vergleichsoperation
       immer das erste der beiden Elemente und wird vor dem anderen Element
       eingefgt. Wird der Vergleich nie wahr, wird das neue Element an
       das Ende der Liste angefgt. *)
  VAR
     neu, hilf: tZeiger;
  BEGIN
    ErzeugeNeuknoten(neu,Adr);
    WITH Liste^ DO
         aktuell:= anker^.nachfolger;
         IF aktuell = anker
             THEN
                  nacheinfuegen(neu,anker)
             ELSE
               WHILE NOT (Vergleich(neu^.info,aktuell^.info) OR
                          (aktuell^.nachfolger = anker))
                   DO aktuell:= aktuell^.nachfolger  END;(* of WHILE *)
               IF Vergleich(neu^.info,aktuell^.info)
                  THEN voreinfuegen(neu,aktuell)
                  ELSE nacheinfuegen(neu,aktuell)  END;
            END;  (* of IF aktuell = anker *)
     END; (* of WITH *);
   END SortiertEinfuegen;

(*************** Positionieren innerhalb der Liste  ********************)

PROCEDURE ZumnaechstenElement(VAR Liste: tList);
    (* Positioniert den aktuellen Zeiger auf das folgende Listenelement.
       Steht der aktuelle Zeiger am Listenende, bleibt er dort stehen. *)
    BEGIN
      IF Liste^.aktuell^.nachfolger <> Liste^.anker  THEN
      Liste^.aktuell := Liste^.aktuell^.nachfolger  END;
    END ZumnaechstenElement;


PROCEDURE ZumvorigenElement(VAR Liste: tList);
 (* Positioniert den aktuellen Zeiger auf das vorangegangene Listenelement.
    Steht der aktuelle Zeiger am Listenanfang, bleibt er dort stehen. *)
  BEGIN
     IF Liste^.aktuell^.vorgaenger <> Liste^.anker  THEN
     Liste^.aktuell := Liste^.aktuell^.vorgaenger   END;
  END ZumvorigenElement;


PROCEDURE ZumVergleichselement(VAR Liste:tList; Adr: ADDRESS;
                               istgleich:VergleicheProc;
                               VAR vorhanden: BOOLEAN);
    (* Positioniert den aktuellen Zeiger auf das Listenelement, dessen
       Schlssel mit einem ueren Schlssel bereinstimmt. Ist das
       Element nicht in der Liste enthalten, wird vorhanden auf FALSE
       gesetzt. (sonst auf TRUE) *)
    BEGIN
      IF leereListe(Liste)  THEN vorhanden := FALSE
      ELSE
      WITH Liste^ DO
        aktuell:= anker^.nachfolger;
        WHILE NOT (istgleich(Adr, aktuell^.info) OR
                   (aktuell^.nachfolger = anker))
            DO aktuell:= aktuell^.nachfolger    END;
        IF istgleich(Adr,aktuell^.info)  THEN vorhanden := TRUE
                                         ELSE vorhanden := FALSE END;
     END; (* of WITH *)
     END; (* IF leereListe *)
    END ZumVergleichselement;


PROCEDURE ZumKopf(VAR Liste:tList);
    (* Positioniert den aktuellen Zeiger auf den Kopf der Liste. *)
    BEGIN
      Liste^.aktuell := Liste^.anker^.nachfolger
    END ZumKopf;


PROCEDURE ZumSchwanz(VAR Liste:tList);
    (* Positioniert den aktuellen Zeiger auf den Schwanz der Liste. *)
    BEGIN
      Liste^.aktuell := Liste^.anker^.vorgaenger
    END ZumSchwanz;


(***************** Ausgabe- und Loeschoperationen ********************)

PROCEDURE LoescheElement(VAR Liste:tList);
    (* Loescht das Element, auf das der aktuelle Zeiger zeigt. *)
    VAR
       hilf, hilf1: tZeiger;
    BEGIN
      WITH Liste^ DO
        IF NOT leereListe(Liste)
        THEN
          DISPOSE(aktuell^.info);
          hilf := aktuell^.vorgaenger;
          hilf1:= aktuell^.nachfolger;
          hilf^.nachfolger := aktuell^.nachfolger;
          hilf1^.vorgaenger:= aktuell^.vorgaenger;
          DISPOSE(aktuell);
        END; (* of IF *)
      END; (* of WITH *)
   END LoescheElement;


PROCEDURE AusgabeElement(Liste:tList):ADDRESS;
    (* Gibt den Zeiger auf den Informationsteil des aktuellen
       Zeigers aus. *)
    BEGIN
      RETURN Liste^.aktuell^.info;
    END AusgabeElement;


PROCEDURE EntferneAusListe(VAR Liste:tList):ADDRESS;
    VAR
       hilf, hilf1: tZeiger;
    BEGIN
      WITH Liste^ DO
        IF NOT leereListe(Liste)
        THEN
          hilf := aktuell^.vorgaenger;
          hilf1:= aktuell^.nachfolger;
          hilf^.nachfolger := aktuell^.nachfolger;
          hilf1^.vorgaenger:= aktuell^.vorgaenger;
          RETURN aktuell^.info;
          DISPOSE(aktuell)
        ELSE RETURN NIL
        END; (* of IF *)
      END; (* of WITH *)
    END EntferneAusListe;

(******************** Positionsabfragen  **********************)

PROCEDURE erstesElement( Liste:tList):BOOLEAN;
    (* wird auf TRUE gesetzt, wenn der aktuelle Zeiger auf den
       Kopf der Liste zeigt. *)
  BEGIN
    IF Liste^.aktuell^.vorgaenger = Liste^.anker  THEN RETURN TRUE
                                                  ELSE RETURN FALSE  END;
  END erstesElement;

PROCEDURE letztesElement(Liste:tList):BOOLEAN;
    (* wird auf TRUE gesetzt, wenn der aktuelle Zeiger auf den
       Schwanz der Liste zeigt. *)
   BEGIN
     IF Liste^.aktuell^.nachfolger = Liste^.anker  THEN RETURN TRUE
                                                   ELSE RETURN FALSE  END;
  END letztesElement;

PROCEDURE leereListe(Liste:tList):BOOLEAN;
    (* wird auf TRUE gesetzt, wenn die Liste keine Eintrge enthlt. *)
    BEGIN
    IF Liste^.anker^.nachfolger = Liste^.anker  THEN RETURN TRUE
                                                ELSE RETURN FALSE  END;
    END leereListe;

(************************************************************************)

END Liste7.